#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>


typedef struct {
		size_t n;
		size_t p;
		double * y_data;
		double * y_calc_data;
		double * x_data;
		double Result_array[Row_max][Col_max];
		double Param_array[Param_max][2];
		int Parameter_no;
	   }Data;

typedef struct {
			int n;
			int p;
			double *x_data;
			double *y_calc_data;
		} Data_nongsl;

int func_calc(double *param, void *calc_data);
void print_state (size_t iter, gsl_multifit_fdfsolver * s, void *data);
int expb_df (const gsl_vector * x, void *data, gsl_matrix * J);
int expb_fdf (const gsl_vector * x, void *data, gsl_vector * f, gsl_matrix * J);
int expb_f (const gsl_vector * x, void *data, gsl_vector * f);


#define FIT(i) gsl_vector_get(s->x, i)
#define ERR(i) sqrt(gsl_matrix_get(covar,i,i))


int LVM_algorithm (Data_struct *Data_structure_ptr)
{
	const gsl_multifit_fdfsolver_type *T;
	gsl_multifit_fdfsolver *s;
	int status;
	unsigned int i,j,k, iter = 0;

	

	const size_t n = Data_structure_ptr->Row_no;
	double x_data[n],y_data[n],y_calc_data[n];
	Data d;

	d.n = n;
	d.y_data=&y_data[0];
	d.x_data=&x_data[0];
	d.y_calc_data=&y_calc_data[0];

	for(i=0;i<n;i++)
	{
		x_data[i] = Data_structure_ptr->Data_array[i][0]; 
		y_data[i] = Data_structure_ptr->Data_array[i][1];
	}



	// Test printing
	for (i = 0; i < n; i++)
	{
		printf ("data: %u\t%lf\t%lf\n", i, x_data[i], y_data[i]);
	};



	// Parameter file



	const size_t p = Data_structure_ptr->Parameter_no;
	double x_init[p];
	d.p = p;
	for(i=0;i<p;i++)
	{
		x_init[i] = Data_structure_ptr->Limit_array[i]; 

	}

	//for (i = 0; i < p; i++)
	//{
	//	printf ("param initials: %u\t%lf\n", i, x_init[i]);
	//};
	//getchar();

	gsl_matrix *covar = gsl_matrix_alloc (p, p);
	gsl_multifit_function_fdf f;




	gsl_vector_view x = gsl_vector_view_array (x_init, p);
	const gsl_rng_type * type;
	gsl_rng * r;
	gsl_rng_env_setup();
	type = gsl_rng_default;
	r = gsl_rng_alloc (type);
	f.f = &expb_f;
	f.df = &expb_df;
	f.fdf = &expb_fdf;
	f.n = n;
	f.p = p;
	f.params = &d;


	T = gsl_multifit_fdfsolver_lmsder;
	s = gsl_multifit_fdfsolver_alloc (T, n, p);
	gsl_multifit_fdfsolver_set (s, &f, &x.vector);
	print_state (iter, s, &d);

	do
	{
		iter++;
		status = gsl_multifit_fdfsolver_iterate (s);
		printf ("status = %s\n", gsl_strerror (status));
		print_state (iter, s, &d);
		if (status)
		break;
		status = gsl_multifit_test_delta (s->dx, s->x, 1e-7, 1e-7);
	}
	while (status == GSL_CONTINUE && iter < Data_structure_ptr->Iterations);



	gsl_multifit_covar (s->J, pow(10,-5), covar);
	double chi = gsl_blas_dnrm2(s->f);
	Data_structure_ptr->Chi_square = chi;
	double dof = n - p;
	double c = chi / sqrt(dof);
	printf("chisq/dof = %g\n", pow(chi, 2.0) / dof);


	for(i=0;i<p;i++)
	{
		printf ("Param[%d] = %.5f +/- %.5f\n",i+1, FIT(i), c*ERR(i));
		Data_structure_ptr->Parameter_array[i][0] = FIT(i);
		Data_structure_ptr->Parameter_array[i][1] = c*ERR(i);
	}


	printf ("status = %s\n", gsl_strerror (status));

	for(i=0;i<n;i++)
	{	
		Data_structure_ptr->Result_array[i][0] = d.Result_array[i][0];
		Data_structure_ptr->Result_array[i][1] = d.Result_array[i][1];
		Data_structure_ptr->Result_array[i][2] = d.Result_array[i][2];
	}

	for(i=0;i<p;i++)
	{	for(j=0;j<p;j++){printf("C[%d,%d]:%lf  ",i,j,gsl_matrix_get(covar,i,j));}
		printf("\n");}

	for(i=0;i<p;i++)
	{	for(j=0;j<p;j++){printf("J[%d,%d]:%lf  ",i,j,gsl_matrix_get(s->J,i,j));}printf("\n");}



	gsl_multifit_fdfsolver_free (s);
	gsl_matrix_free (covar);
	gsl_rng_free (r);
	return 0;
}




void print_state (size_t iter, gsl_multifit_fdfsolver * s, void *data)
{
	int i,j;
	size_t p = ((Data *)data)->p;
	printf("iter: %3u\t",(unsigned int) iter);
	for(i=0;i<p;i++)
	{  printf ( "x[%d] = % 15.8f\t",i+1, gsl_vector_get (s->x, i));} 
 	printf("|f(x)| = %g\n", gsl_blas_dnrm2 (s->f));
}



/* expfit.c -- model functions for exponential + background */

int expb_f (const gsl_vector * x, void *data, gsl_vector * f)
{
	int i,j;
	size_t n = ((Data *)data)->n;
	size_t p = ((Data *)data)->p;
	double *y_data = ((Data *)data)->y_data;
	double *x_data = ((Data *)data)->x_data;
	double Param[p];
	double X_expt,Y_expt,Y_calc;
	for(i=0;i<p;i++)
	{ Param[i] = gsl_vector_get (x, i);}



	Data_nongsl *data_nongsl;
	data_nongsl=(Data_nongsl*)malloc(sizeof(Data_nongsl));
	data_nongsl->n=((Data *)data)->n;
	data_nongsl->p=((Data *)data)->p;
	data_nongsl->x_data = ((Data *)data)->x_data;
	data_nongsl->y_calc_data = ((Data *)data)->y_calc_data;
	func_calc(Param, data_nongsl);


	for (i = 0; i < n; i++)
	{
		/* Model Yi = A * exp(-lambda * i) + b */
		X_expt = x_data[i];
		Y_expt = y_data[i];
		Y_calc = data_nongsl->y_calc_data[i];	// 
		((Data *)data)->Result_array[i][0]=X_expt;
		((Data *)data)->Result_array[i][1]=Y_expt;
		((Data *)data)->Result_array[i][2]=Y_calc;


		gsl_vector_set (f, i, (Y_calc - Y_expt));
	}
return GSL_SUCCESS;
}


int expb_df (const gsl_vector * x, void *data, gsl_matrix * J)
{
	int i,j;
	size_t n = ((Data *)data)->n;
	size_t p = ((Data *)data)->p;
	double *y_data = ((Data *)data)->y_data;
	double *x_data = ((Data *)data)->x_data;
	double Param[p],Param_d[p],diff;
	double Y_calc[n],Y_dcalc[n],Jacobian,dividor;
	diff=pow(10,-3);
	for(i=0;i<p;i++) { Param_d[i] = Param[i] = gsl_vector_get (x, i);}


	Data_nongsl *data_nongsl;
	data_nongsl=(Data_nongsl*)malloc(sizeof(Data_nongsl));
	data_nongsl->n=((Data *)data)->n;
	data_nongsl->p=((Data *)data)->p;
	data_nongsl->x_data = ((Data *)data)->x_data;
	data_nongsl->y_calc_data = ((Data *)data)->y_calc_data;




		// Calculation of the Jacobian matrix column wise

		for(i=0;i<p;i++)
		{	
			Param_d[i] = Param[i]+fabs(Param[i]*diff);

			func_calc(Param, data_nongsl);
			for(j=0;j<n;j++) {Y_calc[j]=data_nongsl->y_calc_data[j];}

			func_calc(Param_d, data_nongsl);
			for(j=0;j<n;j++) {Y_dcalc[j]=data_nongsl->y_calc_data[j];}


			for(j=0;j<n;j++)
			{
				if(Param[i]==0){dividor=pow(10,-5);}
				else{dividor=fabs(Param[i]*diff);}
				Jacobian=(Y_dcalc[j]-Y_calc[j])/dividor;
				gsl_matrix_set (J, j, i, Jacobian);
			}

			for(j=0;j<p;j++){Param_d[j] = Param[j];}
	
		}

	return GSL_SUCCESS;
}


int expb_fdf (const gsl_vector * x, void *data, gsl_vector * f, gsl_matrix * J)
{
	expb_f (x, data, f);
	expb_df (x, data, J);
	return GSL_SUCCESS;
}


int func_calc(double *param, void *calc_data)
{
	int n = ((Data_nongsl *)calc_data)->n;
	int p = ((Data_nongsl *)calc_data)->p;
	double *x_data = ((Data_nongsl *)calc_data)->x_data;
	
		int i,j;


		for(i=0;i<n;i++)
		{
			for(j=0;j<p;j++){if(param[j]==0){param[j]=pow(10,-10);}}
			((Data_nongsl *)calc_data)->y_calc_data[i]=((param[0]*exp(-param[1] * x_data[i]))+param[2]);
			//printf("x:%lf  y:%lf  [%lf  %lf  %lf]\n",x_data[i],((Data_nongsl *)calc_data)->y_calc_data[i],param[0],param[1],param[2]);
		} 

}

